from vasppy.poscar import Poscar
import re
import copy
import numpy as np

class Xdatcar:

    lines_offset = 9

    def __init__( self ):
        """
        Initialise a Xdatcar object.

        Args:
            None

        Returns:
            None
        """
        self.poscar = []
        self.poscar.append( Poscar() )
  
    def read_from( self, filename ):
        self.poscar[ 0 ].read_from( filename )
        with open( filename ) as f:
            lines = f.read()
        frame_header = re.compile( '\nDirect configuration=\s+\d+\n' )
        frame_coordinates = [ frame.split("\n") for frame in frame_header.split( lines )[2:] ]
        for f, frame in enumerate( frame_coordinates ):
            self.poscar.append( copy.deepcopy( self.poscar[ 0 ] ) ) 
            self.poscar[-1].coordinates = np.array( [ [ float( e ) for e in frame.pop(0).split()[0:3] ] for i in range( sum( self.poscar[0].atom_numbers ) ) ] )




import numpy as np

class Rdf:
    """
    class for radial distribution functions
    """

    def __init__( self, max_r, number_of_bins ):
        """
        Initialise a Rdf object for manipulating radial distribution functions.
        
        Args:
            max_r (Float): the maximum r value stored for g(r).
            number_of_bins (Int): number of bins for storing data about g(r).

        Returns:
            None
        """
        self.max_r = max_r
        self.number_of_bins = number_of_bins
        self.data = np.zeros( number_of_bins )
        self.dr = max_r / number_of_bins
 
    def add_dr( self, dr ):
        """
        Add an observed interatomic distance to the g(r) data at dr.

        Args:
            dr (Float): the interatomic distance, dr.

        Returns:
            None
        """ 
        this_bin = int( dr / self.dr ) 
        if this_bin > self.number_of_bins:
            raise IndexError( 'dr is larger than rdf max_r' )
        self.data[ this_bin ] += 1

    def normalised_data( self ):
        return( np.array( [ [ dr, g_of_r / self.volume_factor( dr ) ] for dr, g_of_r in zip( np.array( [ self.dr * ( r + 0.5 ) for r in range( 0, self.number_of_bins ) ] ), self.data / np.sum( self.data ) ) ] ) )  

    def __add__( self, other_rdf ):
        assert isinstance( other_rdf, Rdf )
        assert self.max_r == other_rdf.max_r
        assert self.number_of_bins == other_rdf.number_of_bins
        summed_rdf = Rdf( self.max_r, self.number_of_bins )
        summed_rdf.data = self.data + other_rdf.data
        return summed_rdf 

    def volume_factor( self, dr ):
        return ( ( dr + self.dr )**3 - dr**3 )



#! /usr/bin/env python3

from vasppy.xdatcar import Xdatcar
from vasppy.rdf import Rdf
import argparse
import copy
import math

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument( 'xdatcar' )
    parser.add_argument( 'label', nargs = 2 )
    parser.add_argument( 'max_r', type = float )
    parser.add_argument( 'n_bins', type = int )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    max_r = args.max_r
    number_of_bins = args.n_bins
    species_1 = args.label[ 0 ]
    species_2 = args.label[ 1 ]
    xdatcar = Xdatcar()
    xdatcar.read_from( args.xdatcar )
    rdf = Rdf( max_r = max_r, number_of_bins = number_of_bins )
    volume_scaling_factor = 4.0 * math.pi / ( 3.0 * xdatcar.poscar[0].cell.volume())
    for poscar in xdatcar.poscar:
        rdf += poscar.to_configuration().partial_rdf( species_1, species_2, max_r = max_r, number_of_bins = number_of_bins )
    print( rdf.normalised_data() )
    #[ print( dr, g_of_r / volume_scaling_factor ) for dr, g_of_r in rdf.normalised_data() ]

if __name__ == "__main__":
    main()



#! /usr/bin/env python3 

from vasppy.poscar import Poscar
import string
import argparse      

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser( description='Manipulates VASP POSCAR files' )
    parser.add_argument( 'poscar', help="filename of the VASP POSCAR to be processed" )
    parser.add_argument( '-l', '--label', type=int, choices=[ 1, 4 ], help="label coordinates with atom name at position {1,4}" )
    parser.add_argument( '-c', '--coordinates-only', help='only output coordinates', action='store_true' )
    parser.add_argument( '-t', '--coordinate-type', type=str, choices=[ 'c', 'cartesian', 'd', 'direct' ], default='direct', help="specify coordinate type for output {(c)artesian|(d)irect} [default = (d)irect]" )
    parser.add_argument( '-g', '--group', help='group atoms within supercell', action='store_true' )
    parser.add_argument( '-s', '--supercell', type=int, nargs=3, metavar=( 'h', 'k', 'l' ), help='construct supercell by replicating (h,k,l) times along [a b c]' )
    parser.add_argument( '-b', '--bohr', action='store_true', help='assumes the input file is in Angstrom, and converts everything to bohr')
    parser.add_argument( '-n', '--number-atoms', action='store_true', help='label coordinates with atom number' )
    parser.add_argument( '--scale', action='store_true', help='scale the lattice parameters by the scaling factor' )
    parser.add_argument( '--selective', choices=[ 'T', 'F' ], help='generate Selective Dynamics POSCAR with all values set to T / F' )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    coordinate_types = { 'd' : 'Direct',
                         'direct' : 'Direct',
                         'c' : 'Cartesian',
                         'cartesian' : 'Cartesian' }
    coordinate_type = coordinate_types[ args.coordinate_type ]
    # initialise
    poscar = Poscar()
    # read POSCAR file
    poscar.read_from( args.poscar )
    if args.scale:
        poscar.cell.matrix *= poscar.scaling
        poscar.scaling = 1.0
    if args.supercell: # generate supercell
        if args.group:
            # check that if grouping is switched on, we are asking for a supercell that allows a "3D-chequerboard" pattern.
            for (i,axis) in zip( args.supercell, range(3) ):
                if i%2 == 1 and i > 1:
                    raise Exception( "odd supercell expansions != 1 are incompatible with automatic grouping" )
        poscar = poscar.replicate( *args.supercell, group=args.group )
    if args.bohr:
        poscar = poscar.in_bohr()
    # output to stdout
    output_opts = { 'label'    : args.label,
                    'numbered' : args.number_atoms,
                    'coordinates_only' : args.coordinates_only,
                    'selective': args.selective }
    poscar.output( coordinate_type=coordinate_type, opts=output_opts )

if __name__ == "__main__":
    main()



#! /usr/bin/env python3

from vasppy import grid
import argparse

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser( description='z-projection of a VASP (grid format) file' )
    parser.add_argument( 'gridfile', help="filename of the VASP (grid format) file to be processed" )
    parser.add_argument( '-p', '--projection', choices=[ 'x', 'y', 'z' ], help="output averaged projection perpendicular to [x,y,z]" )
    parser.add_argument( '-o', '--orthorhombic', help='map grid points onto an orthorhombic (non-space filling) grid', action = 'store_true' )
    args = parser.parse_args()
    return args

def main():
    args = parse_command_line_arguments()
    vgrid = grid.Grid()
    vgrid.read_from_filename( args.gridfile )
    if args.orthorhombic:
        vgrid = vgrid.interpolate_to_orthorhombic_grid( vgrid.dimensions )
    if args.projection:
        index = grid.Grid.projections[ args.projection ]
        grid_spacing = vgrid.poscar.cell_lengths()[ index ] / vgrid.dimensions[ index ]
        [ print( i * grid_spacing, av ) for i, av in enumerate( vgrid.average( normal_axis_label = args.projection ) ) ]

if __name__ == "__main__":
    main()



import yaml
import re
from collections import Counter

class Calculation:
    """
    class describing a single VASP calculation
    """
    
    def __init__( self, title, energy, stoichiometry ):
        """
        Initialise a Calculation object

        Args:
            title (Str): The title string for this calculation.
            energy (Float): Final energy in eV.
            stoichiometry (Dict{Str:Int}): A dict desribing the calculation stoichiometry,
                e.g. { 'Ti': 1, 'O': 2 }

        Returns:
            None
        """
        self.title = title
        self.energy = energy
        self.stoichiometry = Counter( stoichiometry )
        
    def __mul__( self, scaling ):
        """
        "Multiply" this Calculation by a scaling factor.
        Returns a new Calculation with the same title, but scaled energy and stoichiometry.

        Args:
            scaling (float): The scaling factor.

        Returns:
            (vasppy.Calculation): The scaled Calculation.
        """
        new_calculation = Calculation( title=self.title, energy=self.energy*scaling, stoichiometry=self.scale_stoichiometry( scaling ) )
        return new_calculation
        
    def __truediv__( self, scaling ):
        """
        Implements division by a scaling factor.
        Returns a new Calculation with the same title, but scaled energy and stoichiometry.

        Args:
            scaling (float): The scaling factor.

        Returns:
            (vasppy.Calculation): The scaled Calculation.
        """
        return self * ( 1 / scaling )
        
    def scale_stoichiometry( self, scaling ):
        """
        Scale the Calculation stoichiometry
        Returns the stoichiometry, scaled by the argument scaling.

        Args:
            scaling (float): The scaling factor.

        Returns:
            (Counter(Str:Int)): The scaled stoichiometry as a :obj:`Counter` of ``label: stoichiometry`` pairs
        """ 
        return { k:v*scaling for k,v in self.stoichiometry.items() }
    
def delta_E( reactants, products, check_balance=True ):
    """
    Calculate the change in energy for reactants --> products.
    
    Args:
        reactants (list(:obj:`vasppy.Calculation`)): A `list` of :obj:`vasppy.Calculation` objects. The initial state.
        products  (list(:obj:`vasppy.Calculation`)): A `list` of :obj:`vasppy.Calculation` objects. The final state.
        check_balance (:obj:`bool`, optional): Check that the reaction stoichiometry is balanced. Default is ``True``.

    Returns:
        (float) The change in energy.
    """
    if check_balance:
        if delta_stoichiometry( reactants, products ) != {}:
            raise ValueError( "reaction is not balanced: {}".format( delta_stoichiometry( reactants, products) ) )
    return sum( [ r.energy for r in products ] ) - sum( [ r.energy for r in reactants ] )

def delta_stoichiometry( reactants, products ):
    """
    Calculate the change in stoichiometry for reactants --> products.

    Args:
        reactants (list(:obj:`vasppy.Calculation`): A `list` of :obj:`vasppy.Calculation objects.` The initial state.
        products  (list(:obj:`vasppy.Calculation`): A `list` of :obj:`vasppy.Calculation objects.` The final state.

    Returns:
        (Counter): The change in stoichiometry.
    """ 
    totals = Counter()
    for r in reactants:
        totals.update( ( r * -1.0 ).stoichiometry )
    for p in products:
        totals.update( p.stoichiometry )
    to_return = {}
    for c in totals:
        if totals[c] != 0:
            to_return[c] = totals[c]
    return to_return

def energy_string_to_float( string ):
    """
    Convert a string of a calculation energy, e.g. '-1.2345 eV' to a float.

    Args:
        string (str): The string to convert.
  
    Return
        (float) 
    """
    energy_re = re.compile( "(-?\d+\.\d+)" )
    return float( energy_re.match( string ).group(0) )
    
def import_calculations_from_file( filename ):
    """
    Construct a list of :obj:`Calculation` objects by reading a YAML file.
    Each YAML document should include ``title``, ``stoichiometry``, and ``energy`` fields, e.g.::

        title: my calculation
        stoichiometry:
            - A: 1
            - B: 2
        energy: -0.1234 eV

    Separate calculations should be distinct YAML documents, separated by `---`
    
    Args:
        filename (str): Name of the YAML file to read.

    Returns:
        (dict(vasppy.Calculation)): A dictionary of :obj:`Calculation` objects. For each :obj:`Calculation` object, the ``title`` field from the YAML input is used as the dictionary key.
    """
    calcs = {}
    with open( filename, 'r' ) as stream:
        docs = yaml.load_all( stream, Loader=yaml.SafeLoader )
        for d in docs:
            stoichiometry = Counter()
            for s in d['stoichiometry']:
                stoichiometry.update( s )
            calcs[ d['title'] ] = Calculation( title=d['title'], 
                                               stoichiometry=stoichiometry, 
                                               energy=energy_string_to_float( d['energy'] ) )
    return calcs



import yaml
import os

my_path = os.path.dirname(__file__)

potcar_md5sum_data = {}
potcar_nelect = {}

potcar_sets = [ 'PBE', 'PBE_52', 'PBE_54', 'GGA', 'USPP_GGA', 'LDA', 'LDA_52', 'LDA_54', 'USPP_LDA' ]

for potcar_set in potcar_sets:
    with open( '{}_md5.yaml'.format( os.path.join( my_path, potcar_set ) ), 'r' ) as stream:
        potcar_md5sum_data[ potcar_set ] = yaml.load( stream, Loader=yaml.SafeLoader  )

for potcar_set in potcar_sets:
    with open( '{}_nelect.yaml'.format( os.path.join( my_path, potcar_set ) ), 'r' ) as stream:
        potcar_nelect[ potcar_set ] = yaml.load( stream, Loader=yaml.SafeLoader  )



import math
import numpy as np

def angle( x, y ):
    """
    Calculate the angle between two vectors, in degrees.

    Args:
        x (np.array): one vector.
        y (np.array): the other vector.

    Returns:
        (float):      the angle between x and y in degrees.
    """
    dot = np.dot( x, y )
    x_mod = np.linalg.norm( x )
    y_mod = np.linalg.norm( y )
    cos_angle = dot / ( x_mod * y_mod )
    return np.degrees( np.arccos( cos_angle ) )

def rotation_matrix(axis, theta):
    """
    Return the 3D rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    
    Args:
        axis (np.array): length 3 numpy array defining the axis of rotation.
        theta (float):   rotation angle in radians.

    Returns:
        (np.array):      the corredponding rotation matrix.
    """
    axis = np.asarray( axis )
    theta = np.asarray( theta )
    axis = axis / math.sqrt( np.dot( axis, axis ) )
    a = math.cos( theta / 2 )
    b, c, d = -axis * math.sin( theta / 2 )
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d
    return np.array( [ [ aa+bb-cc-dd, 2*(bc+ad), 2*(bd-ac) ],
                       [ 2*(bc-ad), aa+cc-bb-dd, 2*(cd+ab) ],
                       [ 2*(bd+ac), 2*(cd-ab), aa+dd-bb-cc ] ] )

class Cell:

    def __init__( self, matrix ):
        """
        Initialise a Cell object.

        Args:
            matrix (np.array): 3x3 numpy array containing the cell matrix.

        Returns:
            None
        """
        assert type( matrix ) is np.ndarray
        assert matrix.shape == ( 3, 3 )
        self.matrix = matrix # 3 x 3 numpy Array
        self.inv_matrix = np.linalg.inv( matrix )

    def dr( self, r1, r2, cutoff=None ):
        """
        Calculate the distance between two fractional coordinates in the cell.
        
        Args:
            r1 (np.array): fractional coordinates for position 1.
            r2 (np.array): fractional coordinates for position 2.
            cutoff (optional:Bool): If set, returns None for distances greater than the cutoff. Default None (unset).

        Returns:
            (float): the distance between r1 and r2.
        """
        delta_r_cartesian = ( r1 - r2 ).dot( self.matrix )
        delta_r_squared = sum( delta_r_cartesian**2 )
        if cutoff != None:
            cutoff_squared = cutoff ** 2
            if delta_r_squared > cutoff_squared:
                return None
        return( math.sqrt( delta_r_squared ) )

    def nearest_image( self, origin, point ):
        """
        Find the fractional_coordinates of the nearest periodic image to a point of origin.

        Args:
            origin (np.array): fractional coordinates of the point of origin.
            point  (np.array): fractional coordinates of the other point.

        Returns:
            (np.array): the fractional coordinates of the nearest image of `point` to `origin`.
        """
        return( origin + self.minimum_image( origin, point ) )

    def minimum_image( self, r1, r2 ):
        """
        Find the minimum image vector from point r1 to point r2.

        Args:
            r1 (np.array): fractional coordinates of point r1.
            r2 (np.array): fractional coordinates of point r2.

        Returns:
            (np.array): the fractional coordinate vector from r1 to the nearest image of r2.
        """
        delta_r = r2 - r1
        delta_r = np.array( [ x - math.copysign( 1.0, x ) if abs(x) > 0.5 else x for x in delta_r ] )
        return( delta_r )

    def minimum_image_dr( self, r1, r2, cutoff=None ):
        """
        Calculate the shortest distance between two points in the cell, 
        accounting for periodic boundary conditions.

        Args:
            r1 (np.array): fractional coordinates of point r1.
            r2 (np.array): fractional coordinates of point r2.
            cutoff (:obj: `float`, optional): if set, return zero if the minimum distance is greater than `cutoff`. Defaults to None.

        Returns:
            (float): The distance between r1 and r2.
        """
        delta_r_vector = self.minimum_image( r1, r2 )
        return( self.dr( np.zeros( 3 ), delta_r_vector, cutoff ) )

    def lengths( self ):
        """
        The cell lengths.

        Args:
            None

        Returns:
            (np.array(a,b,c)): The cell lengths.
        """
        return( np.array( [ math.sqrt( sum( row**2 ) ) for row in self.matrix ] ) )

    def angles( self ):
        """
        The cell angles (in degrees).

        Args:
            None

        Returns:
            (list(alpha,beta,gamma)): The cell angles.
        """
        ( a, b, c ) = [ row for row in self.matrix ]
        return [ angle( b, c ), angle( a, c ), angle( a, b ) ]

    def cartesian_to_fractional_coordinates( self, coordinates ):
        """
        Convert a set of Cartesian coordinates to fractional coordinates in the cell.

        Args:
            coordinates (np.array(dim(N,3))): The set of Cartesian coordinates.

        Returns:
            (np.array(dim(N,3))): The corresponding set of fractional coordinates.
        """
        return( coordinates.dot( self.inv_matrix ) )

    def fractional_to_cartesian_coordinates( self, coordinates ):
        """
        Convert a set of fractional coordinates in the cell to Cartesian coordinates.

        Args:
            coordinates (np.array(dim(N,3))): The set of fractional coordinates.

        Returns:
            (np.array(dim(N,3))): The corresponding set of Cartesian coordinates.
        """
        return( coordinates.dot( self.matrix ) )

    def inside_cell( self, r ):
        """
        Given a fractional-coordinate, if this lies outside the cell return the equivalent point inside the cell.

        Args:
            r (np.array): Fractional coordinates of a point (this may be outside the cell boundaries).

        Returns:
            (np.array): Fractional coordinates of an equivalent point, inside the cell boundaries.
        """
        centre = np.array( [ 0.5, 0.5, 0.5 ] )
        new_r = self.nearest_image( centre, r )
        return new_r

    def volume( self ):
        """
        The cell volume.

        Args:
            None

        Returns:
            (float): The cell volume.
        """
        return np.dot( self.matrix[0], np.cross( self.matrix[1], self.matrix[2] ) ) 

    def unit_vectors( self ):
        """
        The unit vectors for the cell vectors.

        Args:
            None

        Returns:
            (np.array): The unit vectors for the cell vectors.
        """
        return( ( self.matrix.transpose() / self.lengths() ).transpose() )

    def rotate( self, axis, theta ):
        self.matrix = np.array( [ np.dot( rotation_matrix(axis,theta), v) for v in self.matrix ] )



from vasppy import atom, cell, rdf
import numpy as np

class Configuration:
    """
    A Configuration object stores a single structure.
    """

    def __init__( self, cell, atoms ):
        self.cell  = cell 
        self.atoms = atoms

    def dr( self, atom1, atom2 ):
        """
        Calculate the distance between two atoms.

        Args:
            atom1 (vasppy.Atom): Atom 1.
            atom2 (vasppy.Atom): Atom 2.

        Returns:
            (float): The distance between Atom 1 and Atom 2.
        """
        return self.cell.dr( atom1.r, atom2.r )

    def minimum_image_dr( self, atom1, atom2, cutoff=None ):
        return self.cell.minimum_image_dr( atom1.r, atom2.r, cutoff=cutoff )

    def interatomic_distances( self, minimum_image_convention = True ):
        return np.array( [ [ self.minimum_image_dr( atom_i, atom_j ) for atom_j in self.atoms ] for atom_i in self.atoms ] )

    def interatomic_distances_for_atom( self, atom1, minimum_image_convention = True ):
        return np.array( [ self.minimum_image_dr( atom1, atom2 ) for atom2 in self.atoms ] )

    def atoms_with_label( self, label ):
        return filter( lambda atom: atom.label == label, self.atoms )        

    def partial_rdf( self, spec_i, spec_j, max_r, number_of_bins ):
        this_rdf = rdf.Rdf( max_r, number_of_bins )
        atoms_i = list( self.atoms_with_label( spec_i ) )
        atoms_j = list( self.atoms_with_label( spec_j ) )
        for atom_i in atoms_i:
            for atom_j in atoms_j:
                if atom_i is atom_j:
                    continue
                dr = self.minimum_image_dr( atom_i, atom_j )
                if dr <= max_r:
                    this_rdf.add_dr( dr )
        return this_rdf 

    def per_atom_rdf( self, spec_i, spec_j, max_r, number_of_bins ):
        rdfs = []
        atoms_i = list( self.atoms_with_label( spec_i ) )
        atoms_j = list( self.atoms_with_label( spec_j ) )
        for atom_i in atoms_i:
            this_rdf = rdf.Rdf( max_r, number_of_bins )
            for atom_j in atoms_j:
                if atom_i is atom_j:
                    continue
                dr = self.minimum_image_dr( atom_i, atom_j )
                try:
                    this_rdf.add_dr( dr )
                except IndexError:
                    pass
                except:
                    raise
            rdfs.append( this_rdf )
        return rdfs




from scipy.constants import physical_constants, angstrom

angstrom_to_bohr = physical_constants['atomic unit of length'][0] / angstrom
ev_to_hartree = physical_constants['electron volt-hartree relationship'][0]



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib._color_data as mcd

tableau_grey = '#bab0ac'

def pdos_column_names( lmax, ispin ):
    if lmax == 2:
        names = [ 's', 'p_y', 'p_z', 'p_x', 'd_xy', 'd_yz', 'd_z2-r2', 'd_xz', 'd_x2-y2' ]
    elif lmax == 3:
        names = [ 's', 'p_y', 'p_z', 'p_x', 'd_xy', 'd_yz', 'd_z2-r2', 'd_xz', 'd_x2-y2',
                  'f_y(3x2-y2)', 'f_xyz', 'f_yz2', 'f_z3', 'f_xz2', 'f_z(x2-y2)', 'f_x(x2-3y2)' ]
    else:
        raise ValueError( 'lmax value not supported' )
    if ispin == 2:
        all_names = []
        for n in names:
            all_names.extend( [ '{}_up'.format(n), '{}_down'.format(n) ] )
    else:
        all_names = names
    all_names.insert( 0, 'energy' )
    return all_names

class Doscar:
    '''
    Contains all the data in a VASP DOSCAR file, and methods for manipulating this.
    '''

    number_of_header_lines = 6

    def __init__( self, filename, ispin=2, lmax=2, lorbit=11, spin_orbit_coupling=False, read_pdos=True, species=None ):
        '''
        Create a Doscar object from a VASP DOSCAR file.

        Args:
            filename (str): Filename of the VASP DOSCAR file to read.
            ispin (optional:int): ISPIN flag. 
                Set to 1 for non-spin-polarised or 2 for spin-polarised calculations.
                Default = 2.
            lmax (optional:int): Maximum l angular momentum. (d=2, f=3). Default = 2.
            lorbit (optional:int): The VASP LORBIT flag. (Default=11).
            spin_orbit_coupling (optional:bool): Spin-orbit coupling (Default=False).
            read_pdos (optional:bool): Set to True to read the atom-projected density of states (Default=True).
            species (optional:list(str)): List of atomic species strings, e.g. [ 'Fe', 'Fe', 'O', 'O', 'O' ].
                Default=None.
        '''
        self.filename = filename
        self.ispin = ispin
        self.lmax = lmax
        self.spin_orbit_coupling = spin_orbit_coupling
        if self.spin_orbit_coupling:
            raise NotImplementedError( 'Spin-orbit coupling is not yet implemented' )
        self.lorbit = lorbit
        self.pdos = None
        self.species = species
        self.read_header()
        self.read_total_dos()
        if read_pdos:
            try:
                self.read_projected_dos()
            except:
                raise
        # if species is set, should check that this is consistent with the number of entries in the
        # projected_dos dataset
        
    @property
    def number_of_channels( self ):
        if self.lorbit == 11:
            return { 2: 9, 3: 16 }[ self.lmax ]
        raise notImplementedError

    def read_header( self ):
        self.header = []
        with open( self.filename, 'r' ) as file_in:
            for i in range( Doscar.number_of_header_lines ):
                self.header.append( file_in.readline() )
        self.process_header()

    def process_header( self ):
        self.number_of_atoms = int( self.header[0].split()[0] )
        self.number_of_data_points = int( self.header[5].split()[2] )
        self.efermi = float( self.header[5].split()[3] )
        
    def read_total_dos( self ): # assumes spin_polarised
        start_to_read = Doscar.number_of_header_lines
        df = pd.read_csv( self.filename, 
                          skiprows=start_to_read, 
                          nrows=self.number_of_data_points,
                          delim_whitespace=True, 
                          names=[ 'energy', 'up', 'down', 'int_up', 'int_down' ],
                          index_col=False )
        self.energy = df.energy.values
        df.drop( 'energy', axis=1 )
        self.tdos = df
        
    def read_atomic_dos_as_df( self, atom_number ): # currently assume spin-polarised, no-SO-coupling, no f-states
        assert atom_number > 0 & atom_number <= self.number_of_atoms
        start_to_read = Doscar.number_of_header_lines + atom_number * ( self.number_of_data_points + 1 )
        df = pd.read_csv( self.filename,
                          skiprows=start_to_read,
                          nrows=self.number_of_data_points,
                          delim_whitespace=True,
                          names=pdos_column_names( lmax=self.lmax, ispin=self.ispin ),
                          index_col=False )
        return df.drop('energy', axis=1)
    
    def read_projected_dos( self ):
        """
        Read the projected density of states data into """
        pdos_list = []
        for i in range( self.number_of_atoms ):
            df = self.read_atomic_dos_as_df( i+1 )
            pdos_list.append( df )
        self.pdos = np.vstack( [ np.array( df ) for df in pdos_list ] ).reshape( 
            self.number_of_atoms, self.number_of_data_points, self.number_of_channels, self.ispin )
        
    def pdos_select( self, atoms=None, spin=None, l=None, m=None ):
        """
        Returns a subset of the projected density of states array.

        Args:
            atoms (int or list(int)): Atom numbers to include in the selection. Atom numbers count from 1. 
                                   Default is to select all atoms.
            spin (str): Select up or down, or both spin channels to include in the selection.
                        Accepted options are 'up', 'down', and 'both'. Default is to select both spins.
            l (str): Select one angular momentum to include in the selectrion.
                     Accepted options are 's', 'p', 'd', and 'f'. Default is to include all l-values.
                     Setting `l` and not setting `m` will return all projections for that angular momentum value.
            m (list(str)): Select one or more m-values. Requires `l` to be set. 
                           The accepted values depend on the value of `l`:
                           `l='s'`: Only one projection. Not set.
                           `l='p'`: One or more of [ 'x', 'y', 'z' ]
                           `l='d'`: One or more of [ 'xy', 'yz', 'z2-r2', 'xz', 'x2-y2' ]
                           `l='f'`: One or more of [ 'y(3x2-y2)', 'xyz', 'yz2', 'z3', 'xz2', 'z(x2-y2)', 'x(x2-3y2)' ]

        Returns:
            np.array: A 4-dimensional numpy array containing the selected pdos values. 
            The array dimensions are [ atom_no, energy_value, lm-projection, spin ]

        """
        valid_m_values = { 's': [],
                           'p': [ 'x', 'y', 'z' ],
                           'd': [ 'xy', 'yz', 'z2-r2', 'xz', 'x2-y2' ],
                           'f': [ 'y(3x2-y2)', 'xyz', 'yz2', 'z3', 'xz2', 'z(x2-y2)', 'x(x2-3y2)' ] }
        if not atoms:
            atom_idx = list(range( self.number_of_atoms ))
        else:
            atom_idx = atoms
        to_return = self.pdos[ atom_idx, :, :, : ]
        if not spin:
            spin_idx = list(range( self.ispin ))
        elif spin is 'up':
            spin_idx = [0]
        elif spin is 'down':
            spin_idx = [1]
        elif spin is 'both':
            spin_idx = [0,1]
        else:
            raise ValueError( "valid spin values are 'up', 'down', and 'both'. The default is 'both'" )
        to_return = to_return[ :, :, :, spin_idx ]
        if not l:
            channel_idx = list(range( self.number_of_channels ))
        elif l == 's':
            channel_idx = [ 0 ]
        elif l == 'p':
            if not m:
                channel_idx = [ 1, 2, 3 ]
            else: # TODO this looks like it should be i+1
                channel_idx = [ i+1 for i, v in enumerate( valid_m_values['p'] ) if v in m ]
        elif l == 'd':
            if not m:
                channel_idx = [ 4, 5, 6, 7, 8 ]
            else: # TODO this looks like it should be i+4
                channel_idx = [ i+4 for i, v in enumerate( valid_m_values['d'] ) if v in m ]
        elif l == 'f':
            if not m:
                channel_idx = [ 9, 10, 11, 12, 13, 14, 15 ]
            else: # TODO this looks like it should be i+9
                channel_idx = [ i+9 for i, v in enumerate( valid_m_values['f'] ) if v in m ]
        else:
            raise ValueError
        return to_return[ :, :, channel_idx, : ]
    
    def pdos_sum( self, atoms=None, spin=None, l=None, m=None ):
        return np.sum( self.pdos_select( atoms=atoms, spin=spin, l=l, m=m ), axis=(0,2,3) )

    def plot_pdos(self, ax=None, to_plot=None, colors=None, 
                  plot_total_dos=True, xrange=None, ymax=None, 
                  scaling=None, split=False, title=None, title_loc='center',
                  labels=True, title_fontsize=16, legend_pos='outside'):
        if not ax:
            fig, ax = plt.subplots(1, 1, figsize=(8.0,3.0))
        else:
            fig = None
        if not colors:
            colors = mcd.TABLEAU_COLORS        
        color_iterator = (c for c in colors)
        
        if not scaling:
            scaling = []
            
        if xrange:
            e_range = (self.energy >= xrange[0]) & (self.energy <= xrange[1])
        else:
            e_range = np.ma.make_mask( self.energy )
            
        auto_ymax = 0.0
            
        if not to_plot:
            to_plot = {}
            for s in set( self.species ):
                to_plot[s] = ['s', 'p', 'd']
                if self.lmax == 3:
                    to_plot[s].append('f')
                    
        for species in to_plot.keys():
            index = [i for i, s in enumerate(self.species) if s is species]
            for state in to_plot[species]:
                assert state in ['s', 'p', 'd', 'f']
                color = next( color_iterator )
                label = '{} {}'.format(species, state)
                up_dos = self.pdos_sum(atoms=index, l=state, spin='up')[e_range]
                down_dos = self.pdos_sum(atoms=index, l=state, spin='down')[e_range]
                if species in scaling:
                    if state in scaling[species]:
                        up_dos *= scaling[species][state]
                        down_dos *= scaling[species][state]
                        label = r'{} {} $\times${}'.format( species, state, scaling[species][state] )
                auto_ymax = max( [ auto_ymax, up_dos.max(), down_dos.max() ] )
                ax.plot(self.energy[e_range], up_dos, label=label, c=color)
                ax.plot(self.energy[e_range], down_dos * -1.0,  c=color)
        if plot_total_dos:
            ax.fill_between(self.energy[e_range], self.tdos.up.values[e_range], 
                            self.tdos.down.values[e_range] * -1.0, facecolor=tableau_grey, alpha=0.2)
            auto_ymax = max( [ auto_ymax, self.tdos.up.values[e_range].max(), self.tdos.down.values[e_range].max() ] )
    
        if xrange:
            ax.set_xlim( xrange[0], xrange[1] )
            
        if not ymax:
            ymax = 1.1 * auto_ymax
        ax.set_ylim(-ymax*1.1,ymax*1.1)
        if legend_pos == 'outside':
            ax.legend(bbox_to_anchor=(1.01, 1.04), loc='upper left')
        else:
            ax.legend( loc=legend_pos )
        if labels:
            ax.set_xlabel( 'Energy [eV]')
        ax.axhline(y=0, c='lightgrey')
        ax.axes.grid( False, axis='y' )
    
        ax.tick_params(
            axis='y',          # changes apply to the y-axis
            which='both',      # both major and minor ticks are affected
            left=False,      # ticks along the left edge are off
            right=False,         # ticks along the right edge are off
            labelleft=False) # labels along the left edge are off
        
        if title:
            ax.set_title( title, loc=title_loc, fontdict={'fontsize': title_fontsize} )
    
        return fig



#! /usr/bin/env python3 

from vasppy.poscar import Poscar
import argparse
import copy
import numpy as np

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser( description='Reorders the atomic species in a VASP POSCAR file' )
    parser.add_argument( 'labels', nargs='+', help="atomic lables in the desired order" )
    parser.add_argument( 'poscar', help="filename of the VASP POSCAR to be processed" )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    # initialise
    poscar = Poscar()
    # read POSCAR file
    poscar.read_from( args.poscar )
    # construct new Poscar instance with sorted atom types
    sorted_poscar = copy.deepcopy( poscar )
    sorted_poscar.atoms = []
    sorted_poscar.atom_numbers = []
    coordinate_list = []
    atoms = poscar.to_configuration().atoms
    for label in args.labels:
        if label in poscar.atoms:
            sorted_poscar.atoms.append( label )
            matched_atoms = [ atom for atom in atoms if atom.label == label ]
            sorted_poscar.atom_numbers.append( len( matched_atoms ) )
            coordinate_list.extend( [ atom.r for atom in matched_atoms ] )
        else:
            raise( ValueError( "'{}' atom label not found in {}".format( label, args.poscar ) ) )
    sorted_poscar.coordinates = np.array( coordinate_list )
    sorted_poscar.output( coordinate_type = poscar.coordinate_type )

if __name__ == "__main__":
    main()



"""
functions for working with optical properties from vasprun.xml
"""

from math import pi, sqrt
import numpy as np
from scipy.constants import physical_constants, speed_of_light

eV_to_recip_cm = 1.0/(physical_constants['Planck constant in eV s'][0]*speed_of_light*1e2)

def matrix_eigvals(matrix):
    """
    Calculate the eigenvalues of a matrix.

    Args:
        matrix (np.array): The matrix to diagonalise.

    Returns:
        (np.array): Array of the matrix eigenvalues.
    """
    eigvals, eigvecs = np.linalg.eig(matrix)
    return eigvals
       
def to_matrix( xx, yy, zz, xy, yz, xz ):
    """
    Convert a list of matrix components to a symmetric 3x3 matrix.
    Inputs should be in the order xx, yy, zz, xy, yz, xz.

    Args:
        xx (float): xx component of the matrix.
        yy (float): yy component of the matrix.
        zz (float): zz component of the matrix.
        xy (float): xy component of the matrix.
        yz (float): yz component of the matrix.
        xz (float): xz component of the matrix.

    Returns:
        (np.array): The matrix, as a 3x3 numpy array.
    """
    matrix = np.array( [[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]] )
    return matrix

def plot_dielectric_functions( dielectric, ax=None ):
    real_dielectric = parse_dielectric_data( dielectric[1] )
    imag_dielectric = parse_dielectric_data( dielectric[2] )
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(6.0,3.0))
    else:
        fig = None
    ax.plot( dielectric[0], np.mean( real_dielectric, axis=1 ), '-', zorder=2 ) # better to pass in v.dielectric
    ax.plot( dielectric[0], np.mean( imag_dielectric, axis=1 ), '-', zorder=2 )
    ax.set_xlim([0,8])
    ax.set_ylim([0,5])
    return fig

def parse_dielectric_data( data ):
    """
    Convert a set of 2D vasprun formatted dielectric data to
    the eigenvalues of each corresponding 3x3 symmetric numpy matrices.

    Args:
        data (list): length N list of dielectric data. Each entry should be
                     a list of ``[xx, yy, zz, xy, xz, yz ]`` dielectric
                     tensor elements.

    Returns:
        (np.array):  a Nx3 numpy array. Each row contains the eigenvalues
                     for the corresponding row in `data`.
    """ 
    return np.array( [ matrix_eigvals( to_matrix( *e ) ) for e in data ] )

def absorption_coefficient( dielectric ):
    """
    Calculate the optical absorption coefficient from an input set of
    pymatgen vasprun dielectric constant data.

    Args:
        dielectric (list): A list containing the dielectric response function
                           in the pymatgen vasprun format.

                           | element 0: list of energies
                           | element 1: real dielectric tensors, in ``[xx, yy, zz, xy, xz, yz]`` format.
                           | element 2: imaginary dielectric tensors, in ``[xx, yy, zz, xy, xz, yz]`` format.
    
    Returns:
        (np.array): absorption coefficient using eV as frequency units (cm^-1).

    Notes:
        The absorption coefficient is calculated as

        .. math:: \\alpha = \\frac{2\sqrt{2} \pi}{\lambda} \sqrt{-\epsilon_1+\sqrt{\epsilon_1^2+\epsilon_2^2}}

    """
    energies_in_eV = np.array( dielectric[0] )
    real_dielectric = parse_dielectric_data( dielectric[1] )
    imag_dielectric = parse_dielectric_data( dielectric[2] )
    epsilon_1 = np.mean( real_dielectric, axis=1 )
    epsilon_2 = np.mean( imag_dielectric, axis=1 )
    return ( 2.0 * np.sqrt(2.0)*pi*eV_to_recip_cm*energies_in_eV
                 * np.sqrt( -epsilon_1 + np.sqrt( epsilon_1**2 + epsilon_2**2 ) ) )



#! /usr/bin/env python3

# Adapted from http://kitchingroup.cheme.cmu.edu/blog/2013/02/18/Nonlinear-curve-fitting/

import glob
import numpy as np
import pandas as pd
from scipy.optimize import leastsq
import argparse
import warnings

warnings.filterwarnings("ignore", category=UserWarning,
                            module="pymatgen")
from pymatgen.io.vasp import Vasprun
from pymatgen.io.vasp.outputs import UnconvergedVASPWarning

import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt

from vasppy.poscar import Poscar
from vasppy.summary import find_vasp_calculations
from vasppy.utils import match_filename

def parse_args():
    parser = argparse.ArgumentParser(description='Perform a Murnaghan equation of state fit across VASP subdirectories')
    parser.add_argument( '-p', '--plot', action='store_true', help='generate murn.pdf plot of fit' )
    parser.add_argument( '-v', '--verbose', action='store_true', help='verbose output' )
    args = parser.parse_args()
    return args

def read_vasprun( filename ):
    return Vasprun( filename, parse_potcar_file=False, parse_dos=False, parse_eigen=False )

def read_data( verbose=True ):
    dir_list = find_vasp_calculations()
    if not dir_list:
        raise ValueError( 'Did not find any subdirectories containing vasprun.xml or vasprun.xml.gz files' )
    data = []
    for d in dir_list:
        converged = True
        try:
            with warnings.catch_warnings(record=True) as w:
                vasprun = read_vasprun( match_filename( d + 'vasprun.xml' ) )
                for warning in w:
                    if isinstance( warning.message, UnconvergedVASPWarning ):
                        converged = False
                    else:
                        print( warning.message )
        except:
            continue 
        poscar = Poscar.from_file( d + 'POSCAR' )
        data.append( [ poscar.scaling, 
                       vasprun.final_structure.volume, 
                       vasprun.final_energy,
                       converged ] )
    column_titles = [ 'scaling', 'volume', 'energy', 'converged' ]
    df = pd.DataFrame( data, columns=column_titles ).sort_values( by='scaling' )
    df = df.reset_index( drop=True )
    df['scaling_factor'] = df.volume / df.scaling**3
    scaling_factor_round = 4
    if verbose:
        print( df.to_string(index=False) )
    if len( set( df.scaling_factor.round( scaling_factor_round ) ) ) != 1:
        raise ValueError( "POSCAR scaling factors and volumes are inconsistent" )
    return df

def murnaghan( vol, e0, b0, bp, v0 ):
    """
    Calculate the energy as a function of volume, using the Murnaghan equation of state
    [Murnaghan, Proc. Nat. Acad. Sci. 30, 244 (1944)]
    https://en.wikipedia.org/wiki/Murnaghan_equation_of_state
    cf. Fu and Ho, Phys. Rev. B 28, 5480 (1983).

    Args:
        vol (float): this volume.
        e0 (float):  energy at the minimum-energy volume, E0.
        b0 (float):  bulk modulus at the minimum-energy volume, B0.
        bp (float):  pressure-derivative of the bulk modulus at the minimum-energy volume, B0'.
        v0 (float):  volume at the minimum-energy volume, V0.
        
    Returns:
        (float): The energy at this volume. 
    """
    energy = e0 + b0 * vol / bp * (((v0 / vol)**bp) / (bp - 1) + 1) - v0 * b0 / (bp - 1.0)
    return energy

def objective( pars, x, y ):
    err =  y - murnaghan( x, *pars )
    return err

def lstsq_fit( volumes, energies ):
    e_min = energies.min()
    v_min = volumes[ np.argwhere( energies == e_min )[0][0] ]
    x0 = [ e_min, 2.0, 10.0, v_min ] #initial guess of parameters
    plsq = leastsq( objective, x0, args=( volumes, energies ) )
    return plsq

def make_plot( df, fit_params ):
    v_min = df.volume.min()*0.99
    v_max = df.volume.max()*1.01
    v_fitting = np.linspace( v_min, v_max, num=50 )
    e_fitting = murnaghan( v_fitting, *fit_params )
    plt.figure( figsize=(8.0,6.0) )
    # plot converged data points
    loc = df.converged
    plt.plot( df[loc].volume, df[loc].energy, 'o' )
    # plot unconverged data points
    loc = [ not b for b in df.converged ]
    plt.plot( df[loc].volume, df[loc].energy, 'o', c='grey' )
    # plot fitted equation of state curve
    plt.plot( v_fitting, e_fitting, '--' )
    plt.xlabel( r'volume [$\mathrm{\AA}^3$]' )
    plt.ylabel( r'energy [eV]' )
    plt.tight_layout()
    plt.savefig( 'murn.pdf' )

def fit( verbose=False, plot=False ):
    df = read_data( verbose=verbose )
    e0, b0, bp, v0 = lstsq_fit( np.array( df.volume ), np.array( df.energy  ) )[0]
    if plot:
        make_plot( df, ( e0, b0, bp, v0 ) )
    print( "E0: {:.4f}".format( e0 ) )
    print( "V0: {:.4f}".format( v0 ) )
    print( "opt. scaling: {:.5f}".format( ( v0 / df.scaling_factor.mean() )**(1/3) ) )
    
def main():
    args = parse_args()
    fit( verbose=args.verbose, plot=args.plot )

if __name__ == '__main__':
    main()



#! /usr/bin/env python3

import numpy as np
import argparse
from vasppy.poscar import Poscar
from vasppy.cell import Cell
import vasppy.pimaim

def parse_command_line_arguments():
    parser = argparse.ArgumentParser( description = 'TODO' )
    parser.add_argument( '-f', '--filename', help="PIMAIM restart file filename" )
    parser.add_argument( '-l', '--labels', nargs='+', help="labels for each species", required = True )
    parser.add_argument( '-n', '--atom-numbers', nargs='+', help="atom numbers for each species", required = True, type = int )
    args = parser.parse_args()
    if len( args.labels ) != len( args.atom_numbers ):
        print( '\'labels\' and \'atom-numbers\' require matched argument numbers' )
        exit()
    return args 

def main():
    args = parse_command_line_arguments()
    if args.filename is None:
        args.filename = 'testout.rst'
    poscar = vasppy.pimaim.poscar_from_pimaim_restart( args.filename, args.atom_numbers, args.labels )
    poscar.output_as_xtl()

if __name__ == '__main__':
    main()



#! /usr/bin/env python3 

from vasppy.poscar import Poscar
import argparse

def parse_command_line_arguments():
    # command line arguments
    parser = argparse.ArgumentParser( description='Converts a VASP POSCAR file to the .xtl file format' )
    parser.add_argument( 'poscar', help="filename of the VASP POSCAR to be processed" )
    parser.add_argument( '-s', '--symprec', type=float, help="Symmetry precision for a symmetrised .cif output" )
    args = parser.parse_args()
    return( args )

def main():
    args = parse_command_line_arguments()
    # initialise
    poscar = Poscar()
    # read POSCAR file
    poscar.read_from( args.poscar )
    poscar.output_as_cif( args.symprec )

if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import sys
from pathlib import Path
import yaml

"""
Script for collecting information about VASP calculations into YAML format, for further processing.
Expects a series of directories (listed in `result_dirs`) that each contain:
    vasprun.xml
    vaspmeta.yaml (additional metadata providing information about each calculation)
"""

import argparse
from vasppy.summary import Summary, find_vasp_calculations
from vasppy.vaspmeta import VASPMeta

def get_args():
    parser = argparse.ArgumentParser( description='Summarise a VASP calculation.' )
    parser.add_argument( '-r', '--recursive', help='Recursively analyse directories.', action='store_true' )
    parser.add_argument( '-l', '--list', help="List supported data flags.", action='store_true' )
    parser.add_argument( '-p', '--print', help="Specify data to parse.", nargs='*' )
    parser.add_argument( '-f', '--file', help="Specify a file to read data flags from." )
    parser.add_argument( '-c', '--check', help="Checks whether VASP directories contain vaspmeta.yaml and vasprun.xml files", action='store_true' )
    args = parser.parse_args()
    return args

# This should really be set in the vasppy.Summary code, so that it can be tested to be consistent with the supported print methods.
# In fact, ideally the key, print method, and description would all be collected in a single object, which suggests writing this as a simple class.

def main():
    supported_flags = Summary.supported_flags
    to_print=[ 'title', 'status', 'stoichiometry', 'potcar', 'plus_u', 'energy', 'lreal', 'k-points', 'functional', 'encut', 'ediffg', 'ibrion', 'converged', 'version', 'md5', 'directory' ]
    titles = None
    args = get_args()
    if args.list:
        for k, v, in supported_flags.items():
            print( "{}: {}".format( k.ljust(15), v ) )
        sys.exit()
    if args.file:
        with open( args.file, 'r' ) as stream:
            settings = yaml.load( stream, Loader=yaml.SafeLoader ) 
        if 'to_print' in settings:
            to_print = settings['to_print']
        if 'titles' in settings:
            titles = settings['titles']
    if args.print:
        not_supported = [ p for p in args.print if p not in supported_flags ]
        if not_supported:
            raise ValueError( not_supported )
        else:
            to_print = args.print
    if args.recursive:
        path = sorted( find_vasp_calculations() )
    else:
        path = [ '.' ]
    if args.check:
        for p in path:
           vaspmeta = Path( '{}/vaspmeta.yaml'.format( p ) )
           if not vaspmeta.is_file(): 
               print( '{} is missing vaspmeta.yaml'.format( p ) ) 
           vasprun  = Path( '{}/vasprun.xml'.format( p ) )
           if not vasprun.is_file():
               print( '{} is missing vasprun.xml'.format( p ) ) 
    else:
        if titles:
            # Only parse directories with matching vasp_meta titles
            matching_path = []
            for p in path:
                vm = VASPMeta.from_file( '{}/vaspmeta.yaml'.format(p) )
                if vm.title in titles:
                    matching_path.append( p )
            path = matching_path
        for p in path:
            s = Summary( p )
            s.output( to_print=to_print )

if __name__ == "__main__":
    main()



"""
vasppy: Python utilities for working with VASP inputs and outputs
"""

from setuptools import setup, find_packages
from vasppy import __version__ as VERSION

readme = 'README.md'
long_description = open( readme ).read()

scripts = [ 'check_species',
            'murnfit', 
            'vasp_summary', 
            'poscar_to_cif', 
            'potcar_spec',
            'effective_mass',
            'fat_bands',
            'pimaim_to_poscar',
            'pimaim_to_xtl',
            'poscar_sort',
            'poscar_to_pimaim',
            'poscar_to_xtl',
            'proc_poscar',
            'rotate_poscar',
            'spacegroup',
            'vasp_grid',
            'xdatcar_to_disp',
            'xdatcar_to_poscart',
            'xdatcar_to_rdf' ]

setup(
    name='vasppy',
    version=VERSION,
    description='Python utilities for working with VASP inputs and outputs',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Benjamin J. Morgan',
    author_email='bjm42@bath.ac.uk',
    url='https://github.com/bjmorgan/vasppy', 
    download_url='https://github.com/bjmorgan/vasppy/archive/{}.tar.gz'.format( VERSION ),
    keywords=['vasp'], # keywords
    packages=find_packages( exclude=['docs', 'tests*'] ),
    package_data={ 'vasppy': ['data/*.yaml'] },
    entry_points={ 'console_scripts': [
                       '{} = vasppy.scripts.{}:main'.format( s, s ) for s in scripts ] },
    license='MIT',
    install_requires=[ 'monty',
                       'numpy',
                       'pandas',
                       'pymatgen',
                       'PyYAML', 
                       'coverage==4.3.4',
                       'codeclimate-test-reporter',
                       'fortranformat' ],
    python_requires='>=3.5'
    )



# routines for interacting with PIMAIM data / filetypes

import numpy as np
from vasppy.poscar import Poscar
from vasppy.cell import Cell
from .units import angstrom_to_bohr

def lines_to_numpy_array( data ):
    return np.array( [ [ float( s ) for s in line.split() ] for line in data ] )

def read_restart_file( filename, number_of_atoms ):
    with open( filename, 'r' ) as f:
        file_data = f.readlines()

    cr_dump_log, vel_dump_log, chg_dump_log, full_dump_log = [ ( line.strip() == 'T' ) for line in file_data[:4] ]
    # this assumes coordinates, velocities, and dipoles are all present.
    # not sure what happens if atoms have qudrupoles, etc.
    coordinates = lines_to_numpy_array( file_data[ 4 : 4 + number_of_atoms ] ) * angstrom_to_bohr
    if vel_dump_log:
        velocities = lines_to_numpy_array( file_data[ 4 + number_of_atoms : 4 + number_of_atoms * 2 ] )
    else:
        velocities = None
    if chg_dump_log:
        dipoles = lines_to_numpy_array( file_data[ 4 + number_of_atoms * 2 : 4 + number_of_atoms * 3 ] )
    else:
        dipoles = None
    cell_matrix = lines_to_numpy_array( file_data[ -6: -3 ] )
    cell_lengths = lines_to_numpy_array( file_data[ -3: ] ) * angstrom_to_bohr
    full_cell_matrix = cell_matrix * cell_lengths
    
    return( coordinates, velocities, dipoles, full_cell_matrix, cell_lengths )

def get_cart_coords_from_pimaim_restart(coordinates, full_cell_matrix, cell_lengths):

    return(np.dot(coordinates, np.array([full_cell_matrix[i]/cell_lengths[i] for i in range(3)])
))

def poscar_from_pimaim_restart( filename, atom_numbers, atom_labels ):
    number_of_atoms = sum( atom_numbers )
    coordinates, velocities, dipoles, full_cell_matrix, cell_lengths = read_restart_file( filename, number_of_atoms, cell_lengths )

    poscar = Poscar()
    full_cell_matrix = full_cell_matrix.transpose()
    coordinates = get_cart_coords_from_pimaim_restart(coordinates, full_cell_matrix,cell_lengths)
    poscar.cell = Cell( full_cell_matrix)
    poscar.atoms = atom_labels
    poscar.atom_numbers = atom_numbers
    poscar.coordinate_type = 'Direct'
    poscar.coordinates = poscar.cell.cartesian_to_fractional_coordinates( coordinates )
 
    return poscar



#! /usr/bin/env python3

from vasppy import procar
from vasppy.outcar import reciprocal_lattice_from_outcar
import argparse

def minimum_length( nmin ):
    class MinimumLength( argparse.Action ):
        def __call__( self, parser, args, values, option_string=None ):
            if not nmin <= len( values ):
                msg = 'argument "{f}" requires at least {nmin} arguments'.format( f = self.dest, nmin = nmin )
                raise argparse.ArgumentError( self, msg )
            setattr( args, self.dest, values )
    return MinimumLength
            
def main():
    parser = argparse.ArgumentParser( description='Calculate an effective mass from a VASP PROCAR using a fitted quadratic' )
    parser.add_argument( '-k', '--k-points', help='index of k-points for calculating effective mass', nargs='+', type=int, required=True, action=minimum_length( 2 ) )
    parser.add_argument( '-b', '--band-index', help='index of band for calculating effective mass', type=int, required=True )
    parser.add_argument( '-f', '--procar', help='PROCAR filename (default PROCAR)', type=str, default='PROCAR' )
    parser.add_argument( '-v', '--verbose', help='Verbose output', action='store_true' )
    parser.add_argument( '-o', '--outcar', help='OUTCAR filename (default OUTCAR)', type=str, default='OUTCAR' )
    parser.add_argument( '-s', '--spin', help='select spin channel (default 1 / non-spin-polarised)', type=int, default='1' )
    args = parser.parse_args()

    reciprocal_lattice = reciprocal_lattice_from_outcar( 'OUTCAR' ) # Move reading the reciprocal lattice to procar.py

    pcar = procar.Procar()
    pcar.read_from_file( args.procar )
    effective_mass = pcar.effective_mass_calc( k_point_indices = args.k_points, 
                                               band_index = args.band_index, 
                                               reciprocal_lattice = reciprocal_lattice,
                                               spin = args.spin,
                                               printing = args.verbose )
    print( effective_mass )

if __name__ == '__main__':
    main()



#! /usr/bin/env python3

from vasppy.summary import potcar_spec
import argparse

def parse_command_line_arguments():
    parser = argparse.ArgumentParser( description='Generate POTCAR specification based on hashing individual pseudopotential strings' )
    parser.add_argument( 'potcar', help="filename of the VASP POTCAR to be processed", nargs='?', default='POTCAR' )
    args = parser.parse_args()
    return args

def main():
    args = parse_command_line_arguments()
    for p, ps in potcar_spec( args.potcar ).items():
        print( p, ps )

if __name__ == '__main__':
    main()



import numpy as np
import re
import math
import warnings
from .units import angstrom_to_bohr, ev_to_hartree
from copy import deepcopy
import fortranformat as ff

class KPoint():

    def __init__( self, index, frac_coords, weight ):
        self.index = index
        self.frac_coords = frac_coords
        self.weight = weight

    def cart_coords( self, reciprocal_lattice ):
        """Convert the reciprocal fractional coordinates for this k-point to \
        reciprocal Cartesian coordinates.

        Args:
            reciprocal_lattice (np.array(float)): 3x3 numpy array containing the \
                Cartesian reciprocal lattice.

        Returns:
            (np.array): The reciprocal Cartesian coordinates of this k-point.

        """
        return np.dot( self.frac_coords, reciprocal_lattice )

    def __eq__( self, other ):
        return ( ( self.index == other.index ) &
                 ( self.frac_coords == other.frac_coords ).all() &
                 ( self.weight == other.weight ) )

    def __repr__( self ):
        return "k-point {}: {} weight = {}".format( self.index, ' '.join( [ str(c) for c in self.frac_coords ] ), self.weight )

def handle_occupancy( occupancy, negative_occupancies='warn' ):
    valid_negative_occupancies = [ 'warn', 'raise', 'ignore', 'zero' ]
    if negative_occupancies not in valid_negative_occupancies:
        raise ValueError( "valid options for negative_occupancies are {}".format( valid_negative_occupancies ) )
    if occupancy < 0:
       if negative_occupancies == 'warn':
           warnings.warn( "One or more occupancies in your PROCAR file are negative." )
       elif negative_occupancies == 'raise':
           raise ValueError( "One or more occupancies in your PROCAR file are negative." )
       elif negative_occupancies == 'ignore':
           pass
       elif negative_occupancies == 'zero':
           occupancy = 0.0
    return occupancy

class Band():

    def __init__( self, index, energy, occupancy, negative_occupancies='warn' ):
        self.index = index
        self.energy = energy
        self.occupancy = handle_occupancy( occupancy, negative_occupancies=negative_occupancies )

    def __eq__( self, other ):
        return ( ( self.index == other.index ) & 
                 ( self.energy == other.energy ) & 
                 ( self.occupancy == other.occupancy ) )

def get_numbers_from_string( string ):
    p = re.compile('-?\d+[.\d]*')
    return [ float( s ) for s in p.findall( string ) ]

def k_point_parser( string ):
    """Parse k-point data from a PROCAR string.

    Finds all lines of the form::

         k-point    1 :    0.50000000 0.25000000 0.75000000     weight = 0.00806452

    and extracts the k-point index, reciprocal fractional coordinates, and weight
    into a :obj:`procar.KPoint` object.

    Args:
        string (str): String containing a full PROCAR file.

    Returns:
        (list(:obj:`procar.KPoint`)): A list of :obj:`procar.KPoint` objects.

    """
    regex = re.compile( 'k-point\s+(\d+)\s*:\s+([- ][01].\d{8})([- ][01].\d{8})([- ][01].\d{8})\s+weight = ([01].\d+)' )
    captured = regex.findall( string ) 
    k_points = []
    for kp in captured:
        index = int(kp[0])
        frac_coords = np.array( [ float(s) for s in kp[1:4] ] )
        weight = float(kp[4])
        k_points.append( KPoint( index=index, frac_coords=frac_coords, weight=weight ) )
    return k_points

def projections_parser( string ):
    regex = re.compile( '([-.\d\se]+tot.+)\n' )
    data = regex.findall( string )
    data = [ x.replace( 'tot', '0' ) for x in data ]
    data = np.array( [ x.split() for x in data ], dtype = float )
    return data

def area_of_a_triangle_in_cartesian_space( a, b, c ):
    """
    Returns the area of a triangle defined by three points in Cartesian space.

    Args:
        a (np.array): Cartesian coordinates of point A.
        b (np.array): Cartesian coordinates of point B.
        c (np.array): Cartesian coordinates of point C.

    Returns:
        (float): the area of the triangle.
    """
    return 0.5 * np.linalg.norm( np.cross( b-a, c-a ) )

def points_are_in_a_straight_line( points, tolerance=1e-7 ):
    """
    Check whether a set of points fall on a straight line.
    Calculates the areas of triangles formed by triplets of the points.
    Returns False is any of these areas are larger than the tolerance.

    Args:
        points (list(np.array)): list of Cartesian coordinates for each point.
        tolerance (optional:float): the maximum triangle size for these points to be considered colinear. Default is 1e-7.

    Returns:
        (bool): True if all points fall on a straight line (within the allowed tolerance).
    """
    a = points[0]
    b = points[1]
    for c in points[2:]:
        if area_of_a_triangle_in_cartesian_space( a, b, c ) > tolerance:
            return False
    return True

def two_point_effective_mass( cartesian_k_points, eigenvalues ):
    """
    Calculate the effective mass given eigenvalues at two k-points.
    Reimplemented from Aron Walsh's original effective mass Fortran code.

    Args:
        cartesian_k_points (np.array): 2D numpy array containing the k-points in (reciprocal) Cartesian coordinates.
        eigenvalues (np.array):        numpy array containing the eigenvalues at each k-point.

    Returns:
        (float): The effective mass
    """
    assert( cartesian_k_points.shape[0] == 2 )
    assert( eigenvalues.size == 2 )
    dk = cartesian_k_points[ 1 ] - cartesian_k_points[ 0 ]
    mod_dk = np.sqrt( np.dot( dk, dk ) )
    delta_e = ( eigenvalues[ 1 ] - eigenvalues[ 0 ] ) * ev_to_hartree * 2.0
    effective_mass = mod_dk * mod_dk / delta_e
    return effective_mass

def least_squares_effective_mass( cartesian_k_points, eigenvalues ):
    """
    Calculate the effective mass using a least squares quadratic fit.

    Args:
        cartesian_k_points (np.array): Cartesian reciprocal coordinates for the k-points
        eigenvalues (np.array):        Energy eigenvalues at each k-point to be used in the fit.

    Returns:
        (float): The fitted effective mass

    Notes:
        If the k-points do not sit on a straight line a ValueError will be raised.
    """
    if not points_are_in_a_straight_line( cartesian_k_points ):
        raise ValueError( 'k-points are not collinear' )
    dk = cartesian_k_points - cartesian_k_points[0]
    mod_dk = np.linalg.norm( dk, axis = 1 )
    delta_e = eigenvalues - eigenvalues[0]
    effective_mass = 1.0 / ( np.polyfit( mod_dk, eigenvalues, 2 )[0] * ev_to_hartree * 2.0 )
    return effective_mass

class Procar:
    """
    Object for working with PROCAR data.

    Attributes:
        data (numpy.array(float)): A 5D numpy array that stores the projection data.

                    Axes are k-points, bands, spin-channels, ions and sum over ions, lm-projections.

        bands (numpy.array(float)): A list of ``Band`` objects, that contain band index, energy, and occupancy data.
        k_points (list(KPoint)): A list of ``KPoint`` objects, that contain fractional coordinates and weights for each k-point.
        number_of_k_points (int): The number of k-points.
        number_of_bands (int): The number of bands.
        spin_channels (int): Number of spin channels in the PROCAR data:
 
                    -  1 for non-spin-polarised calculations.
                    -  2 for spin-polarised calculations.
                    -  4 for non-collinear calculations.
        number_of_ions (int): The number of ions.
        number_of_projections (int): The number of projections, e.g. TODO
        calculation (dict): Dictionary of True | False values describing the calculation type.

            Dictionary keys are 'non_spin_polarised', 'non_collinear', and 'spin_polarised'.
      
    """

    def __init__( self, spin=1, negative_occupancies='warn' ):
        self._spin_channels = spin # should be determined from PROCAR
        self._number_of_k_points = None
        self._number_of_ions = None
        self._number_of_bands = None
        self._number_of_projections = None
        self._k_point_blocks = None
        self.data = None
        self.bands = None
        self.calculation = { 'non_spin_polarised': False, 'non_collinear': False, 'spin_polarised': False }
        if negative_occupancies not in [ 'warn', 'raise', 'zero' ]:
            raise ValueError( "negative_occupancies can be one of [ 'warn', 'raise', 'zero' ]" )
        self.negative_occupancies = negative_occupancies
        #self.non_spin_polarised = None

    @property
    def occupancy( self ):
        return np.array( [ [ band.index, band.occupancy ] for band in self.bands ] )

    def __add__( self, other ):
        if self.spin_channels != other.spin_channels:
            raise ValueError( 'Can only concatenate Procars with equal spin_channels: {}, {}'.format( self.spin_channels, other.spin_channels ) )
        if self.number_of_ions != other.number_of_ions:
            raise ValueError( 'Can only concatenate Procars with equal number_of_ions: {}, {}'.format( self.number_of_ions, other.number_of_ions ) )
        if self.number_of_bands != other.number_of_bands:
            raise ValueError( 'Can only concatenate Procars with equal number_of_bands: {}, {}'.format( self.number_of_bands, other.number_of_bands ) )
        if self.number_of_projections != other.number_of_projections:
            raise ValueError( 'Can only concatenate Procars with equal number_of_projections: {}, {}'.format( self.number_of_projections, other.number_of_projections ) )
        if self._k_point_blocks != other._k_point_blocks:
            raise ValueError( 'Can only concatenate Procars with equal k_point_blocks: {}, {}'.format( self._k_point_blocks, other._k_point_blocks ) )
        if self.calculation != other.calculation:
            raise ValueError( 'Can only concatenate Procars from equal calculations: {}, {}'.format( self.calculation, other.calculation ) )
        new_procar = deepcopy( self )
        new_procar.data = np.concatenate( ( self.data, other.data ) )
        new_procar._number_of_k_points = self.number_of_k_points + other.number_of_k_points
        new_procar.bands = deepcopy(self.bands) + deepcopy(other.bands)
        new_procar.k_points = deepcopy(self.k_points) + deepcopy(other.k_points)
        for i, kp in enumerate( new_procar.k_points, 1 ):
            kp.index = i    
        new_procar.sanity_check()
        return new_procar
 
    def parse_projections( self ):
        self.projection_data = projections_parser( self.read_in )
        try:
            assert( self._number_of_bands * self._number_of_k_points == len( self.projection_data ) )
            self._spin_channels = 1 # non-magnetic, non-spin-polarised
            self._k_point_blocks = 1
            self.calculation[ 'non_spin_polarised' ] = True
        except:
            if self._number_of_bands * self._number_of_k_points * 4 == len( self.projection_data ):
                self._spin_channels = 4 # non-collinear (spin-orbit coupling)
                self._k_point_blocks = 1
                self.calculation[ 'non_collinear' ] = True
                pass
            elif self._number_of_bands * self._number_of_k_points * 2 == len( self.projection_data ):
                self._spin_channels = 2 # spin-polarised
                self._k_point_blocks = 2
                self.calculation[ 'spin_polarised' ] = True
                pass
            else:
                raise
        self._number_of_projections = int( self.projection_data.shape[1] / ( self._number_of_ions + 1 ) ) - 1

    def parse_k_points( self ):
        self.k_points = k_point_parser( self.read_in )[:self._number_of_k_points]

    def parse_bands( self ):
        band_data = re.findall( r"band\s*(\d+)\s*#\s*energy\s*([-.\d]+)\s?\s*#\s"r"*occ.\s*([-.\d]+)", self.read_in )
        self.bands = [ Band( float(i), float(e), float(o), negative_occupancies=self.negative_occupancies ) for i, e, o in band_data ]

    def sanity_check( self ):
        assert( self._number_of_k_points == len( self.k_points ) ), "k-point number mismatch: {} in header; {} in file".format( self._number_of_k_points, len( self.k_points ) )
        read_bands = len( self.bands ) / self._number_of_k_points / self._k_point_blocks
        assert( self._number_of_bands == read_bands ), "band mismatch: {} in header; {} in file".format( self._number_of_bands, read_bands )

    @classmethod
    def from_files( cls, filenames, negative_occupancies='warn' ):
        """Create a :obj:`Procar` object by reading the projected wavefunction character of each band
        from a series of VASP ``PROCAR`` files.

        Useful when e.g. a band-structure calculation has been split over multiple VASP calculations,
        for example, when using hybrid functionals.

        Args:
            filename (str): Filename of the ``PROCAR`` file.
            negative_occupancies (:obj:`Str`, optional): Select how negative occupancies are handled.
                Options are:

                    - ``warn`` (default): Warn that some partial occupancies are negative.
                    - ``raise``:          Raise an AttributeError.
                    - ``ignore``:         Do nothing.
                    - ``zero``:           Negative partial occupancies will be set to zero.

        Returns:
            (:obj:`vasppy.Procar`)
        
        """
        pcars = [ cls.from_file( f, negative_occupancies=negative_occupancies ) for f in filenames ]
        return sum( pcars )

    @classmethod
    def from_file( cls, filename, negative_occupancies='warn' ):
        """Create a :obj:`Procar` object by reading the projected wavefunction character of each band
        from a VASP ``PROCAR`` file.

        Args:
            filename (str): Filename of the ``PROCAR`` file.
            negative_occupancies (:obj:`Str`, optional): Select how negative occupancies are handled.
                Options are:

                    - ``warn`` (default): Warn that some partial occupancies are negative.
                    - ``raise``:          Raise an AttributeError.
                    - ``ignore``:         Do nothing.
                    - ``zero``:           Negative partial occupancies will be set to zero.

        Returns:
            (:obj:`vasppy.Procar`)
        
        """
        pcar = cls( negative_occupancies=negative_occupancies )
        pcar.read_from_file( filename=filename )
        return pcar
        
    def read_from_file( self, filename ):
        """Reads the projected wavefunction character of each band from a VASP PROCAR file.

        Args:
            filename (str): Filename of the PROCAR file.

        Returns:
            None
        
        """
        with open( filename, 'r' ) as file_in:
            file_in.readline()
            self._number_of_k_points, self._number_of_bands, self._number_of_ions = [ int( f ) for f in get_numbers_from_string( file_in.readline() ) ]
            self.read_in = file_in.read()
        self.parse_k_points()
        self.parse_bands()
        self.parse_projections()
        self.sanity_check()
        self.read_in = None # clear memory
        if self.calculation[ 'spin_polarised' ]:
            self.data = self.projection_data.reshape( self._spin_channels, self._number_of_k_points, self._number_of_bands, self._number_of_ions+1, self._number_of_projections+1 )[:,:,:,:,1:].swapaxes( 0, 1).swapaxes( 1, 2 )
        else:
            self.data = self.projection_data.reshape( self._number_of_k_points, self._number_of_bands, self._spin_channels, self._number_of_ions+1, self._number_of_projections+1 )[:,:,:,:,1:]

    @property
    def number_of_k_points( self ):
        """The number of k-points described by this :obj:`Procar` object."""
        assert( self._number_of_k_points == self.data.shape[0] ), "Number of k-points in metadata ({}) not equal to number in PROCAR data ({})".format( self._number_of_k_points, self.data.shape[0] )
        return self._number_of_k_points

    @property
    def number_of_bands( self ):
        """The number of bands described by this :obj:`Procar` object."""
        assert( self._number_of_bands == self.data.shape[1] ), "Number of bands in metadata ({}) not equal to number in PROCAR data ({})".format( self._number_of_bands, self.data.shape[1] )
        return self._number_of_bands

    @property
    def spin_channels( self ):
        """The number of spin-channels described by this :obj:`Procar` object."""
        assert( self._spin_channels == self.data.shape[2] ), "Number of spin channels in metadata ({}) not equal to number in PROCAR data ({})".format( self._spin_channels, self.data.shape[2] )
        return self._spin_channels

    @property
    def number_of_ions( self ):
        """The number of ions described by thie :obj:`Procar` object."""
        assert( self._number_of_ions == self.data.shape[3]-1 ), "Number of ions in metadata ({}) not equal to number in PROCAR data ({})".format( self._number_of_ions, self.data.shape[3]-1 )
        return self._number_of_ions

    @property
    def number_of_projections( self ):
        """The number of lm-projections described by this :obj:`Procar` object."""
        assert (self._number_of_projections == self.data.shape[4]), "Number of projections in metadata ({}) not equal to number in PROCAR data ({})".format( self._number_of_projections, self.data.shape[4] ) 
        return self._number_of_projections

    def print_weighted_band_structure( self, spins=None, ions=None, orbitals=None, scaling=1.0, e_fermi=0.0, reciprocal_lattice=None ):
        band_structure_data = self.weighted_band_structure( spins=spins, ions=ions, orbitals=orbitals, scaling=scaling, e_fermi=e_fermi, reciprocal_lattice=reciprocal_lattice ) 
        for i, band_data in enumerate( band_structure_data, 1 ):
            print( '# band: {}'.format( i ) )
            for k_point_data in band_data:
                print( ' '.join( [ str(f) for f in k_point_data ] ) )
            print()

    def weighted_band_structure( self, spins=None, ions=None, orbitals=None, scaling=1.0, e_fermi=0.0, reciprocal_lattice=None ):
        if spins:
            spins = [ s - 1 for s in spins ]
        else:
            spins = list( range( self.spin_channels ) )
        if not ions:
            ions = [ self.number_of_ions ] # nions+1 is the `tot` index
        if not orbitals:
            orbitals = list( range( self.number_of_projections ) )
        if self.calculation[ 'spin_polarised' ]:
            band_energies = np.array( [ band.energy for band in self.bands ] ).reshape( self.spin_channels, self.number_of_k_points, self.number_of_bands )[ spins[0] ].T
        else:
            band_energies = np.array( [ band.energy for band in self.bands ] ).reshape( self.number_of_k_points, self.number_of_bands ).T
        orbital_projection = np.sum( self.data[ :, :, :, :, orbitals ], axis = 4 )
        ion_projection = np.sum( orbital_projection[ :, :, :, ions ], axis = 3 )
        spin_projection = np.sum( ion_projection[ :, :, spins ], axis = 2 )
        x_axis = self.x_axis( reciprocal_lattice )
        to_return = []
        for i in range( self.number_of_bands ):
            for k, ( e, p ) in enumerate( zip( band_energies[i], spin_projection.T[i] ) ):
                to_return.append( [ x_axis[ k ], e - e_fermi, p * scaling ] )
        to_return = np.array( to_return ).reshape( self.number_of_bands, -1, 3 )
        return to_return

    def effective_mass_calc( self, k_point_indices, band_index, reciprocal_lattice, spin=1, printing=False ):
        assert( spin <= self.k_point_blocks )
        assert( len( k_point_indices ) > 1 ) # we need at least 2 k-points
        band_energies = self.bands[:,1:].reshape( self.k_point_blocks, self.number_of_k_points, self.number_of_bands )
        frac_k_point_coords = np.array( [ self.k_points[ k - 1 ].frac_coords for k in k_point_indices ] )
        eigenvalues = np.array( [ band_energies[ spin - 1 ][ k - 1 ][ band_index - 1 ] for k in k_point_indices ] )
        if printing:
            print( '# h k l e' )
            [ print( ' '.join( [ str( f ) for f in row ] ) ) for row in np.concatenate( ( frac_k_point_coords, np.array( [ eigenvalues ] ).T ), axis = 1 ) ]
        reciprocal_lattice = reciprocal_lattice * 2 * math.pi * angstrom_to_bohr
        cart_k_point_coords = np.array( [ k.cart_coords( reciprocal_lattice ) for k in k_points ] ) # convert k-points to cartesian
        if len( k_point_indices ) == 2:
            effective_mass_function = two_point_effective_mass
        else:
            effective_mass_function = least_squares_effective_mass
        return effective_mass_function( cart_k_point_coords, eigenvalues )

    def x_axis( self, reciprocal_lattice=None ):
        """Generate the x-axis values for a band-structure plot.

        Returns an array of cumulative distances in reciprocal space between sequential k-points.

        Args:
            reciprocal_lattice (:obj:`np.array`, optional): 3x3 Cartesian reciprocal lattice.    
                Default is ``None``. If no reciprocal lattice is provided, the returned x-axis
                values will be sequential integers, giving even spacings between sequential
                k-points.
        
        Returns:
            (np.array): An array of x-axis values.
 
        """
        if reciprocal_lattice is not None:
            cartesian_k_points = np.array( [ k.cart_coords( reciprocal_lattice ) for k in k_points ] )
            x_axis = [ 0.0 ]
            for i in range( 1, len( cartesian_k_points ) ):
                dk = cartesian_k_points[ i - 1 ] - cartesian_k_points[ i ]
                mod_dk = np.sqrt( np.dot( dk, dk ) )
                x_axis.append( mod_dk + x_axis[-1] )
            x_axis = np.array( x_axis )
        else:
            x_axis = np.arange( len( self.k_points ) )
        return x_axis

# TODO Need complete set of example PROCAR files before we start to write
# something that can write these all back out in the appropriate format
#    def write_file( self, filename ):
#        if self.calculation['non_collinear']:
#            raise NotImplementedError
#        with open( filename, 'w' ) as f:
#            f.write( 'PROCAR lm decomposed' )
#            for s in range( self.spin_channels ): # not sure what happens for non-collinear calculations
#                line = ff.FortranRecordWriter("/'# of k-points:',I5,9X,'# of bands:',I5,9X,'# of ions:',I5")
#                f.write( line.write( [ self.number_of_k_points, self.number_of_bands, self.number_of_ions ] )+'\n' )
#                for k_point in self.k_points:
#                    line = ff.FortranRecordWriter("/' k-point ',I5,' :',3X,3F11.8,'     weight = ',F10.8/")
#                    f.write( line.write( [ k_point.index, *k_point.frac_coords, k_point.weight ] ) )
#                    for band in self.bands:
#                        line = ff.FortranRecordWriter("/'band ',I5,' # energy',F14.8,' # occ.',F12.8/")
#                        f.write( line.write( [ band.index, band.energy, band.occupancy ] ) )
#                        f.write( '\nion      s     py     pz     px    dxy    dyz    dz2    dxz    dx2    tot\n' )
